from __future__ import annotations

from dataclasses import dataclass
from enum import IntFlag
from typing import Mapping, Self

from little_a2s.events import ClientEventRules
from little_a2s.reader import Reader


@dataclass(kw_only=True)
class Arma3Rules:
    """A deserialized set of rules, mods, and signatures for Arma 3
    and similar Real Virtuality games like DayZ.

    Reference: https://community.bistudio.com/wiki/Arma_3:_ServerBrowserProtocol3

    .. versionadded:: 0.3.0

    .. versionchanged:: 0.7.1

        Fixed handling DayZ payloads with :meth:`from_rules()`.

        If the rules mapping version is 2, the difficulty will always
        be set to ``Arma3Difficulty.from_int(0)``.

    """

    version: int
    """The version of this protocol."""
    overflow: int
    """General flags (overflow flags)."""
    dlc: Arma3DLC
    """A set of DLC flags indicated by the server.

    Some bits may be set that aren't part of this flag's members, up to 0xFFFF.

    """
    difficulty: Arma3Difficulty
    """The difficulty options defined by the server."""
    dlc_hashes: list[int]
    """A list of hashes for each DLC flag set, in order."""
    mods: list[Arma3Mod]
    """A list of mods that are required by the server, including CDLC."""
    signatures: list[str]
    """A list of key filenames loaded on the server, not including the .bikey suffix."""

    @classmethod
    def from_rules(cls, rules: Mapping[bytes, bytes] | ClientEventRules) -> Self:
        """Parse this class from a rules mapping.

        :raises EOFError: Not enough bytes could be read.
        :raises ValueError: The data is malformed.

        """
        if isinstance(rules, ClientEventRules):
            rules = rules.rules

        data = cls._assemble_rules(rules)
        data = cls._translate_escapes(data)
        reader = Reader(data)

        version = reader.read_byte()
        overflow = reader.read_byte()

        if version == 2:
            # https://community.bistudio.com/wiki/Arma_3:_ServerBrowserProtocol2
            # Below works for DayZ, but it doesn't conform to the wiki
            # which says there's meant to be two bytes for difficulty...
            dlc = Arma3DLC(reader.read_ushort())
            difficulty = Arma3Difficulty.from_int(0)
        else:  # version == 3
            dlc = Arma3DLC(reader.read_ushort())
            difficulty = Arma3Difficulty.from_int(reader.read_ushort())

        dlc_hashes = [reader.read_ulong() for _ in range(dlc.bit_count())]

        n_mods = reader.read_byte()
        mods = [Arma3Mod.from_reader(reader) for _ in range(n_mods)]

        n_signatures = reader.read_byte()
        signatures = [reader.read_varchar1().decode() for _ in range(n_signatures)]

        return cls(
            version=version,
            overflow=overflow,
            dlc=dlc,
            difficulty=difficulty,
            dlc_hashes=dlc_hashes,
            mods=mods,
            signatures=signatures,
        )

    @staticmethod
    def _assemble_rules(rules: Mapping[bytes, bytes]) -> bytes:
        # Real Virtuality stores the payload in chunks of 125 bytes,
        # ordered and enumerated.
        data = bytearray()

        last_current = None
        last_total = None

        for k, v in rules.items():
            if len(k) != 2:
                continue

            current, total = k
            if last_total is not None and total != last_total:
                # This could be a legitimate two character key
                raise ValueError(f"Expected total {last_total}, got {total}")
            if last_current is not None and current != last_current + 1:
                raise ValueError(f"Expected index {last_current + 1}, got {current}")
            last_current, last_total = current, total

            data.extend(v)

            # We're making an assumption that the rules are encoded in order,
            # so last key should end with the same index.
            if current == total:
                break

        return bytes(data)

    @staticmethod
    def _translate_escapes(data: bytes) -> bytes:
        # FIXME: does this need to be optimized?
        data = data.replace(b"\x01\x03", b"\xff")
        data = data.replace(b"\x01\x02", b"\x00")
        data = data.replace(b"\x01\x01", b"\x01")
        return data


class Arma3DLC(IntFlag):
    """A set of DLC flags indicated by the server.

    Some bits may be set that aren't part of this flag's members, up to 0xFFFF.

    .. versionadded:: 0.3.0

    """

    KART = 0x1
    MARKSMEN = 0x2
    HELI = 0x4
    CURATOR = 0x8
    EXPANSION = 0x10
    JETS = 0x20
    ORANGE = 0x40  # Laws of War
    ARGO = 0x80
    TACOPS = 0x100
    TANKS = 0x200
    CONTACT = 0x400  # Contact platform
    ENOCH = 0x800  # Contact DLC
    ART_OF_WAR = 0x1000


@dataclass(kw_only=True)
class Arma3Difficulty:
    """The difficulty options defined by the server.

    .. versionadded:: 0.3.0

    """

    difficulty: int
    """The difficulty level, usually between 0-3."""
    skill: int
    """The AI skill level, usually between 0-3."""
    advanced_flight_model: bool
    """Indicates if Advanced Flight Model is enabled."""
    third_person_view: bool
    """Indicates if third-person view is enabled."""
    weapon_crosshair: bool
    """Indicates if the weapon crosshair is enabled."""

    @classmethod
    def from_int(cls, n: int) -> Self:
        """Parse this class from a 16-bit unsigned integer."""
        assert 0 <= n < 0xFFFF
        return cls(
            difficulty=n & 0b111,
            skill=n >> 3 & 0b111,
            advanced_flight_model=bool(n >> 6 & 0b1),
            third_person_view=bool(n >> 7 & 0b1),
            weapon_crosshair=bool(n >> 8 & 0b1),
        )


@dataclass(kw_only=True)
class Arma3Mod:
    """A mod required by the server.

    .. versionadded:: 0.3.0

    """

    hash: int
    """The 32-bit hash of that mod."""
    dlc: bool
    """Indicates if this is a DLC mod."""
    steam_id: int
    """The workshop ID of the mod, or app ID if it is DLC."""
    name: str
    """The name of the mod. Usually omitted for DLC."""

    @classmethod
    def from_reader(cls, reader: Reader) -> Self:
        """Parse this class from a reader.

        :raises EOFError: Not enough bytes could be read.
        :raises ValueError: The data is malformed.

        """
        hash = reader.read_ulong()

        length = reader.read_byte()
        dlc = bool(length >> 4 & 1)
        length = length & 0x0F
        steam_id = int.from_bytes(reader.read(length), "little")
        name = reader.read_varchar1().decode()

        return cls(
            hash=hash,
            dlc=dlc,
            steam_id=steam_id,
            name=name,
        )



from typing import Iterable, Iterator, Type, TypeVar

from little_a2s.events import ClientEvent

Address = tuple[str, int] | tuple[str, int, int, int] | tuple[int, bytes]
T = TypeVar("T")
ClientEventT = TypeVar("ClientEventT", bound=ClientEvent)


def filter_type(
    t: Type[T] | tuple[Type[T], ...],
    it: Iterable[object],
    /,
) -> Iterator[T]:
    """Filter through an iterable for elements of the given type.

    .. versionadded:: 0.2.0

    """
    for x in it:
        if isinstance(x, t):
            yield x


def first(t: Type[T] | tuple[Type[T], ...], it: Iterable[object], /) -> T | None:
    """Return the first element of the given type in an iterable.

    .. versionadded:: 0.2.0

    """
    return next(filter_type(t, it), None)


def last(t: Type[T] | tuple[Type[T], ...], it: Iterable[object], /) -> T | None:
    """Return the last element of the given type in an iterable.

    .. versionadded:: 0.2.0

    """
    x = None
    for x in filter_type(t, it):
        pass
    return x



from collections.abc import Mapping, Sequence
from dataclasses import dataclass
from enum import IntEnum
from typing import Self, SupportsIndex, overload

from little_a2s.enums import _EnumReprMixin
from little_a2s.reader import Reader


# ClientEventInfo types
class ServerType(_EnumReprMixin, IntEnum):
    """Indicates the type of server."""

    DEDICATED = ord("d")
    LISTEN = ord("l")
    RELAY = ord("p")


class Environment(_EnumReprMixin, IntEnum):
    """Indicates the operating system of the server."""

    LINUX = ord("l")
    WINDOWS = ord("w")
    MACOS_M = ord("m")  # Ergh, can this be combined?
    MACOS_O = ord("o")


class Visibility(_EnumReprMixin, IntEnum):
    """Indicates whether the server requires a password."""

    PUBLIC = 0
    PRIVATE = 1


class VAC(_EnumReprMixin, IntEnum):
    """Specifies whether the server uses Valve Anti Cheat."""

    INSECURE = 0
    SECURE = 1


@dataclass(kw_only=True)
class ExtraInfo:
    """Extra data included with an A2S_INFO response."""

    port: int | None = None
    """The server's game port number."""
    steam_id: int | None = None
    """The server's steam ID."""
    spectator_port: int | None = None
    """The spectator port number for SourceTV."""
    spectator_name: str | None = None
    """The name of the spectator server for SourceTV."""
    keywords: str | None = None
    """Tags that describe the game according to the server."""
    game_id: int | None = None
    """The server's game ID."""

    @classmethod
    def from_reader(cls, reader: Reader, flag: int) -> Self:
        """Parse this class from a reader.

        :param reader: The reader to consume.
        :param flag: The Extra Data Flag (EDF) indicating which fields are included.

        """
        extra = cls()
        if flag & 0x80:
            extra.port = reader.read_ushort()
        if flag & 0x10:
            extra.steam_id = reader.read_uint64()
        if flag & 0x40:
            extra.spectator_port = reader.read_ushort()
            extra.spectator_name = reader.read_null_utf8()
        if flag & 0x20:
            extra.keywords = reader.read_null_utf8()
        if flag & 0x01:
            extra.game_id = reader.read_uint64()
        return extra


# ClientEventGoldsourceInfo types
class GoldsourceModType(_EnumReprMixin, IntEnum):
    """Indicates the type of mod for :class:`GoldsourceMod`."""

    SINGLE_AND_MULTIPLAYER = 0
    MULTIPLAYER_ONLY = 1


class GoldsourceModDLL(_EnumReprMixin, IntEnum):
    """Indicates whether a :class:`GoldsourceMod` uses its own DLL."""

    NATIVE = 0
    """This mod uses the Half-Life DLL."""

    EXTENSION = 1
    """This mod provides its own DLL."""


@dataclass(kw_only=True)
class GoldsourceMod:
    """Information about the Goldsource mod, if the game is a mod."""

    link: str
    """URL to mod website. May be empty."""
    download_link: str
    """URL to download the mod. May be empty."""
    version: int
    """Version of mod installed on server."""
    size: int
    """Space (in bytes) the mod takes up."""
    type: GoldsourceModType
    """Indicates the type of mod."""
    dll: GoldsourceModDLL
    """Indicates whether the mod uses its own DLL."""

    @classmethod
    def from_reader(cls, reader: Reader) -> Self:
        """Parse this class from a reader."""
        link = reader.read_null_utf8()
        download_link = reader.read_null_utf8()
        reader.read_null()
        version = reader.read_ulong()
        size = reader.read_ulong()
        type = GoldsourceModType(reader.read_byte())
        dll = GoldsourceModDLL(reader.read_byte())

        return cls(
            link=link,
            download_link=download_link,
            version=version,
            size=size,
            type=type,
            dll=dll,
        )


# ClientEventPlayer types
@dataclass(kw_only=True)
class Player:
    """A player returned in the A2S_PLAYER response."""

    index: int
    """Index of player."""
    name: str
    """Name of the player."""
    score: int
    """Player's score, such as kills."""
    duration: float
    """Time (in seconds) player has been connected to the server."""


# Event types
@dataclass(kw_only=True)
class Event:
    """The base class for all A2S protocol events."""


@dataclass(kw_only=True)
class ClientEvent(Event):
    """An A2S client protocol event."""


@dataclass(kw_only=True)
class ClientEventInfo(ClientEvent):
    """An A2S_INFO client protocol event.

    This follows the Source format. For the Goldsource equivalent,
    see :class:`ClientEventGoldsourceInfo`.

    """

    protocol: int
    """Protocol version used by the server."""
    name: str
    """Name of the server."""
    map: str
    """Map the server has currently loaded."""
    folder: str
    """Name of the folder containing the game files."""
    game: str
    """Full name of the game."""
    id: int
    """Steam Application ID of game."""
    players: int
    """Number of players on the server."""
    max_players: int
    """Maximum number of players the server reports it can hold."""
    bots: int
    """Number of bots on the server. """
    type: ServerType
    """Indicates the type of server."""
    environment: Environment
    """Indicates the operating system of the server."""
    visibility: Visibility
    """Indicates whether the server requires a password."""
    vac: VAC
    """Specifies whether the server uses Valve Anti Cheat."""
    version: str
    """Version of the game installed on the server."""
    extra: ExtraInfo | None
    """Extra data included with the response."""

    @classmethod
    def from_reader(cls, reader: Reader) -> Self:
        """Parse this class from a reader."""
        protocol = reader.read_byte()
        name = reader.read_null_utf8()
        map = reader.read_null_utf8()
        folder = reader.read_null_utf8()
        game = reader.read_null_utf8()
        id = reader.read_ushort()
        players = reader.read_byte()
        max_players = reader.read_byte()
        bots = reader.read_byte()
        type = ServerType(reader.read_byte())
        environment = Environment(reader.read_byte())
        visibility = Visibility(reader.read_byte())
        vac = VAC(reader.read_byte())
        # Extra data will be here for The Ship
        version = reader.read_null_utf8()

        try:
            extra_flag = reader.read_byte()
        except EOFError:
            extra = None
        else:
            extra = ExtraInfo.from_reader(reader, extra_flag)

        return cls(
            protocol=protocol,
            name=name,
            map=map,
            folder=folder,
            game=game,
            id=id,
            players=players,
            max_players=max_players,
            bots=bots,
            type=type,
            environment=environment,
            visibility=visibility,
            vac=vac,
            version=version,
            extra=extra,
        )


@dataclass(kw_only=True)
class ClientEventGoldsourceInfo(ClientEvent):
    """An A2S_INFO Goldsource client protocol event."""

    address: str
    """IP address and port of the server."""
    name: str
    """Name of the server."""
    map: str
    """Map the server has currently loaded."""
    folder: str
    """Name of the folder containing the game files."""
    game: str
    """Full name of the game."""
    players: int
    """Number of players on the server."""
    max_players: int
    """Maximum number of players the server reports it can hold."""
    protocol: int
    """Protocol version used by the server."""
    type: ServerType
    """Indicates the type of server."""
    environment: Environment
    """Indicates the operating system of the server."""
    visibility: Visibility
    """Indicates whether the server requires a password."""
    mod: GoldsourceMod | None
    """Information about the Goldsource mod, if the game is a mod."""
    vac: VAC
    """Specifies whether the server uses Valve Anti Cheat."""
    bots: int
    """Number of bots on the server. """

    @classmethod
    def from_reader(cls, reader: Reader) -> Self:
        """Parse this class from a reader."""
        address = reader.read_null_utf8()
        name = reader.read_null_utf8()
        map = reader.read_null_utf8()
        folder = reader.read_null_utf8()
        game = reader.read_null_utf8()
        players = reader.read_byte()
        max_players = reader.read_byte()
        protocol = reader.read_byte()
        type = ServerType(ord(reader.read_char().lower()))
        environment = Environment(ord(reader.read_char().lower()))
        visibility = Visibility(reader.read_byte())
        mod = GoldsourceMod.from_reader(reader) if reader.read_byte() == 1 else None
        vac = VAC(reader.read_byte())
        bots = reader.read_byte()

        return cls(
            address=address,
            name=name,
            map=map,
            folder=folder,
            game=game,
            players=players,
            max_players=max_players,
            protocol=protocol,
            type=type,
            environment=environment,
            visibility=visibility,
            mod=mod,
            vac=vac,
            bots=bots,
        )


@dataclass(kw_only=True)
class ClientEventPlayers(ClientEvent, Sequence[Player]):
    """An A2S_PLAYER client protocol event.

    .. versionchanged:: 0.5.1

        Added support for the sequence protocol to interact with
        the underlying list of players.

    """

    players: list[Player]
    """List of players whose information was gathered."""

    @classmethod
    def from_reader(cls, reader: Reader) -> Self:
        """Parse this class from a reader."""
        players = []
        for _ in range(reader.read_byte()):
            index = reader.read_byte()
            name = reader.read_null_utf8()
            score = reader.read_long()
            duration = reader.read_float()
            # Extra data will be here for The Ship
            player = Player(index=index, name=name, score=score, duration=duration)
            players.append(player)

        return cls(players=players)

    def __contains__(self, item):
        return item in self.players

    @overload
    def __getitem__(self, index: SupportsIndex) -> Player: ...
    @overload
    def __getitem__(self, index: slice) -> list[Player]: ...
    def __getitem__(self, index):
        return self.players[index]

    def __iter__(self):
        yield from self.players

    def __len__(self):
        return len(self.players)


@dataclass(kw_only=True)
class ClientEventRules(ClientEvent, Mapping[bytes, bytes]):
    """An A2S_RULES client protocol event.

    .. versionchanged:: 0.5.1

        Added support for the mapping protocol to interact with
        the underlying dictionary of rules.

    """

    rules: dict[bytes, bytes]
    """The server rules or configuration variables.

    While the protocol states these should be UTF-8 strings, some games might
    provide binary data in rules which may not decode correctly as UTF-8.
    If you know the game doesn't do this, you can use the :meth:`decode()`
    method for convenience.

    """

    def decode(self) -> dict[str, str]:
        """Return all rules decoded with UTF-8."""
        return {k.decode(): v.decode() for k, v in self.rules.items()}

    @classmethod
    def from_reader(cls, reader: Reader) -> Self:
        """Parse this class from a reader."""
        rules = {}
        for _ in range(reader.read_ushort()):
            name = reader.read_null_string()
            value = reader.read_null_string()
            rules[name] = value

        return cls(rules=rules)

    def __contains__(self, item):
        return item in self.rules

    def __getitem__(self, index):
        return self.rules[index]

    def __iter__(self):
        yield from self.rules

    def __len__(self):
        return len(self.rules)

    def keys(self):
        return self.rules.keys()

    def items(self):
        return self.rules.items()

    def values(self):
        return self.rules.values()


@dataclass(kw_only=True)
class ClientEventChallenge(ClientEvent):
    """An S2C_CHALLENGE client protocol event."""

    challenge: int
    """The challenge number to append to subsequent requests."""

    @classmethod
    def from_reader(cls, reader: Reader) -> Self:
        """Parse this class from a reader."""
        return cls(challenge=reader.read_long())



from .async_ import AsyncA2S as AsyncA2S, AsyncA2SGoldsource as AsyncA2SGoldsource
from .sync import A2S as A2S, A2SGoldsource as A2SGoldsource
from .types import (
    filter_type as filter_type,
    first as first,
    last as last,
)



from __future__ import annotations

import asyncio
import logging
import socket
from contextlib import asynccontextmanager
from functools import partial
from typing import AsyncIterator, Awaitable, Callable, Self, Type

from little_a2s.client.types import Address, ClientEventT
from little_a2s.errors import ChallengeError
from little_a2s.events import (
    ClientEvent,
    ClientEventChallenge,
    ClientEventGoldsourceInfo,
    ClientEventInfo,
    ClientEventPlayers,
    ClientEventRules,
)
from little_a2s.packets import ClientPacket
from little_a2s.protocol import A2SClientProtocol, A2SGoldsourceClientProtocol

log = logging.getLogger(__name__)


class AsyncA2S(asyncio.DatagramProtocol):
    """An asynchronous client for A2S queries.

    ::

        a2s = AsyncA2S.from_addr("127.0.0.1", 27015)
        async with a2s, asyncio.timeout(1):
            print(await a2s.info())
            print(await a2s.players())
            print(await a2s.rules())

    This follows the Source format. For the Goldsource equivalent,
    see :class:`AsyncA2SGoldsource`.

    This class supports the asynchronous context manager protocol which calls
    the connector function to set the transport + remote address, and closes
    the transport upon exit.

    :param connector:
        The function to call and await to create a datagram transport
        and return the remote address, if any.
        See also :meth:`from_addr()`, :meth:`from_ipv4()`, and :meth:`from_ipv6()`.

    .. versionadded:: 0.5.0

    .. versionchanged:: 0.7.0

        Fixed :exc:`PayloadError` not being raised by request methods.

        Concurrent, pending requests to one address can raise the same PayloadError
        if any received packet is malformed.

    """

    _remote_addr: Address | None = None
    _transport: asyncio.DatagramTransport | None = None
    _protocols: dict[Address, A2SClientProtocol]
    _requests: dict[
        tuple[Address, Type[ClientEvent]],
        asyncio.Future[ClientEvent | None],
    ]
    _close_fut: asyncio.Future[None]

    def __init__(
        self,
        connector: Callable[[Self], Awaitable[Address | None]],
    ) -> None:
        self.connector = connector

        self._lock = asyncio.Lock()
        self._request_cond = asyncio.Condition(self._lock)
        self._reset()

    def _reset(self) -> None:
        self._protocols = {}
        self._requests = {}

        loop = asyncio.get_running_loop()
        self._close_fut = loop.create_future()

    # Connection methods

    async def __aenter__(self) -> Self:
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc_val, tb):
        await self.close()
        self._reset()

    async def start(self) -> None:
        """Call the connector function to create the datagram transport.

        :raises OSError: The address could not be resolved.
        :raises RuntimeError: The transport is already connected.

        """
        async with self._lock:
            if self._transport is not None:
                raise RuntimeError("Transport already connected")

            self._remote_addr = await self.connector(self)

            if self._transport is None:
                raise RuntimeError("Connector failed to call connection_made()")

    async def close(self) -> None:
        """Close the current datagram transport, raising any exception
        if the connection improperly closed.

        :raises RuntimeError: The transport is not connected.

        """
        self.transport.close()
        return await asyncio.shield(self._close_fut)

    @property
    def transport(self) -> asyncio.DatagramTransport:
        """The current datagram transport.

        :raises RuntimeError: The transport is not connected.

        """
        if self._transport is None:
            raise RuntimeError("Transport not connected")
        return self._transport

    # Constructor methods

    @classmethod
    def from_addr(
        cls,
        host: str,
        port: int,
        *,
        prefer_ipv4: bool = True,
    ) -> Self:
        """Resolve the given host and create an A2S query.

        :param host: The IPv4 address, IPv6 address, or domain name to query.
        :param port: The port to query.
        :param prefer_ipv4: If True, prefer to resolve hostnames to IPv4 addresses.

        """
        connector = partial(
            cls._connect_from_addr,
            host=host,
            port=port,
            prefer_ipv4=prefer_ipv4,
        )
        return cls(connector)

    async def _connect_from_addr(
        self,
        *,
        host: str,
        port: int,
        prefer_ipv4: bool = True,
    ) -> Address:
        loop = asyncio.get_running_loop()
        addresses = await loop.getaddrinfo(
            host,
            port,
            type=socket.SOCK_DGRAM,
            proto=socket.IPPROTO_UDP,
        )
        if not addresses:
            raise OSError("Address could not be resolved")

        if prefer_ipv4:
            addr = next((a for a in addresses if a[0] == socket.AF_INET), addresses[0])
        else:
            addr = addresses[0]

        family, _, proto, _, addr = addr
        await loop.create_datagram_endpoint(
            lambda: self,
            remote_addr=addr[:2],
            family=family,
            proto=proto,
        )
        return addr

    @classmethod
    def from_ipv4(cls) -> Self:
        """Create an A2S query with a UDP IPv4 socket not connected to any address.

        This allows you to use the same socket with ``addr`` arguments::

            async with AsyncA2S.from_ipv4() as a2s, asyncio.timeout(1):
                info = await a2s.info(("127.0.0.1", 2303))
                info = await a2s.info(("127.0.0.1", 27015))

        """
        connector = partial(cls._connect_from_family, family=socket.AF_INET)
        return cls(connector)

    @classmethod
    def from_ipv6(cls) -> Self:
        """Create an A2S query with a UDP IPv6 socket not connected to any address.

        This allows you to use the same socket with ``addr`` arguments::

            async with AsyncA2S.from_ipv6() as a2s, asyncio.timeout(1):
                info = await a2s.info(("::1", 2303))
                info = await a2s.info(("::1", 27015))

        """
        connector = partial(cls._connect_from_family, family=socket.AF_INET6)
        return cls(connector)

    async def _connect_from_family(self, *, family: socket.AddressFamily) -> None:
        loop = asyncio.get_running_loop()
        await loop.create_datagram_endpoint(
            lambda: self,
            local_addr=("::" if family == socket.AF_INET6 else "0.0.0.0", 0),
            family=family,
            proto=socket.IPPROTO_UDP,
        )

    # Request methods

    async def info(self, addr: Address | None = None) -> ClientEventInfo:
        """Send an A2S_INFO request and wait for a response.

        :param addr:
            The address to send the request to.
            Does not apply if socket is already connected to an address,
            such as from :meth:`from_addr()`.

        :raises ChallengeError: The server sent too many challenges.
        :raises PayloadError: The server sent a malformed packet.
        :raises TypeError: The addr argument was required or forbidden.

        """
        addr = self._get_addr(addr)
        proto = self._get_protocol(addr)
        return await self._send(ClientEventInfo, addr, proto.info)

    async def players(self, addr: Address | None = None) -> ClientEventPlayers:
        """Send an A2S_PLAYER request and wait for a response.

        :param addr:
            The address to send the request to.
            Does not apply if socket is already connected to an address,
            such as from :meth:`from_addr()`.

        :raises ChallengeError: The server sent too many challenges.
        :raises PayloadError: The server sent a malformed packet.
        :raises TypeError: The addr argument was required or forbidden.

        """
        addr = self._get_addr(addr)
        proto = self._get_protocol(addr)
        return await self._send(ClientEventPlayers, addr, proto.players)

    async def rules(self, addr: Address | None = None) -> ClientEventRules:
        """Send an A2S_RULES request and wait for a response.

        :param addr:
            The address to send the request to.
            Does not apply if socket is already connected to an address,
            such as from :meth:`from_addr()`.

        :raises ChallengeError: The server sent too many challenges.
        :raises PayloadError: The server sent a malformed packet.
        :raises TypeError: The addr argument was required or forbidden.

        """
        addr = self._get_addr(addr)
        proto = self._get_protocol(addr)
        return await self._send(ClientEventRules, addr, proto.rules)

    def _get_addr(self, addr: Address | None) -> Address:
        if self._remote_addr and addr:
            raise TypeError("Transport has remote address, addr= is disallowed")
        elif not self._remote_addr and not addr:
            raise TypeError("Transport has no remote address, addr= is required")
        return addr or self._remote_addr  # type: ignore

    def _get_protocol(self, addr: Address) -> A2SClientProtocol:
        """Get the A2S protocol for the given address, creating a new one
        if it doesn't already exist.

        :param addr: The address to bind to.

        """
        proto = self._protocols.get(addr)
        if proto is not None:
            return proto

        proto = self._create_protocol()
        self._protocols[addr] = proto
        return proto

    def _create_protocol(self) -> A2SClientProtocol:
        """Create the A2S protocol to manage state.

        This method can be overridden by subclasses.

        """
        return A2SClientProtocol()

    # DatagramProtocol methods

    def connection_made(self, transport: asyncio.DatagramTransport) -> None:
        self._transport = transport

    def connection_lost(self, exc: Exception | None) -> None:
        if exc is None:
            self._close_fut.set_result(None)
        else:
            self._close_fut.set_exception(exc)

    def datagram_received(self, data: bytes, addr: Address) -> None:
        proto = self._protocols.get(addr)
        if proto is None:
            return log.debug("Ignoring unexpected address %r", addr)

        try:
            proto.receive_datagram(data)
        except Exception as e:  # PayloadError
            return self._set_addr_exception(addr, e)

        events = proto.events_received()
        challenge = None

        for e in events:
            if isinstance(e, ClientEventChallenge):
                challenge = e
                continue

            key = (addr, type(e))
            fut = self._requests.get(key)
            if fut is None or fut.done():
                log.debug("Ignoring unexpected %s", type(e).__name__)
                continue

            fut.set_result(e)

        if challenge is None:
            return

        for (fut_addr, _), fut in self._requests.items():
            if addr == fut_addr and not fut.done():
                fut.set_result(None)

    def _set_addr_exception(self, addr: Address, e: BaseException) -> None:
        for (fut_addr, _), fut in self._requests.items():
            if addr == fut_addr and not fut.done():
                fut.set_exception(e)

    async def _send(
        self,
        t: Type[ClientEventT],
        addr: Address,
        payload: Callable[[], ClientPacket],
    ) -> ClientEventT:
        key = (addr, t)

        for _ in range(3):
            async with self._claim_request(key) as fut:
                self.transport.sendto(bytes(payload()), addr)
                event = await fut

            if event is not None:
                return event

        raise ChallengeError("Server responded with too many challenges")

    @asynccontextmanager
    async def _claim_request(
        self,
        key: tuple[Address, Type[ClientEvent]],
    ) -> AsyncIterator[asyncio.Future]:
        loop = asyncio.get_running_loop()

        async with self._request_cond:
            await self._request_cond.wait_for(lambda: key not in self._requests)
            self._requests[key] = fut = loop.create_future()

        try:
            yield fut
        finally:
            async with self._request_cond:
                self._requests.pop(key, None)
                self._request_cond.notify_all()


class AsyncA2SGoldsource(AsyncA2S):
    """A asynchronous client for A2S Goldsource queries."""

    async def info(self, addr: Address | None = None) -> ClientEventGoldsourceInfo:  # type: ignore
        addr = self._get_addr(addr)
        proto = self._get_protocol(addr)
        return await self._send(ClientEventGoldsourceInfo, addr, proto.info)

    def _create_protocol(self) -> A2SGoldsourceClientProtocol:
        return A2SGoldsourceClientProtocol()



class Error(Exception):
    """The base class for A2S exceptions.

    .. versionadded:: 0.6.0

    """


class PayloadError(ValueError, Error):
    """A malformed A2S packet was received.

    This is raised by :class:`A2SClientProtocol` and propagated up to
    the :class:`A2S` and :class:`AsyncA2S` clients.

    .. versionadded:: 0.6.0

    """


class ChallengeError(Error):
    """The server repeatedly challenged our client in spite of us attempting
    to comply with their challenges.

    This is a rare exception that suggests the server's A2S implementation is broken.

    .. versionadded:: 0.6.0

    """



import logging
import socket
from typing import Callable, Self, Type

from little_a2s.client.constants import DEFAULT_TIMEOUT
from little_a2s.client.types import Address, ClientEventT, filter_type, first
from little_a2s.errors import ChallengeError
from little_a2s.events import (
    ClientEvent,
    ClientEventChallenge,
    ClientEventGoldsourceInfo,
    ClientEventInfo,
    ClientEventPlayers,
    ClientEventRules,
)
from little_a2s.packets import ClientPacket
from little_a2s.protocol import A2SClientProtocol, A2SGoldsourceClientProtocol

log = logging.getLogger(__name__)


class A2S:
    """A synchronous client for A2S queries.

    ::

        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.settimeout(1)
        sock.connect(("127.0.0.1", 27015))
        with A2S(sock) as a2s:
            print(a2s.info())
            print(a2s.players())
            print(a2s.rules())

    This follows the Source format. For the Goldsource equivalent,
    see :class:`A2SGoldsource`.

    This class supports the context manager protocol which automatically
    closes the socket upon exit.

    This class is not thread-safe.

    :param sock:
        The UDP socket to send and receive queries from.
        The socket can be connected to a remote address beforehand
        with :meth:`~socket.socket.connect()`, if you want to skip
        ``addr`` arguments in send methods. You may also want to
        set a timeout with :meth:`~socket.socket.settimeout()`.
        Alternatively, use :meth:`from_addr()`, :meth:`from_ipv4()`,
        or :meth:`from_ipv6()` to construct the socket for you.

    .. versionadded:: 0.2.0

    .. versionchanged:: 0.5.0

        Removed the ``challenge=`` parameter.

    .. versionchanged:: 0.5.0

        Removed the ``events_received()`` method.

    .. versionchanged:: 0.6.0

        Passing a socket that is not of type :data:`socket.SOCK_DGRAM`
        will raise :exc:`TypeError` instead of :exc:`ValueError`.

    """

    buffer_size = 32768  # probably overkill?
    _remote_addr: Address | None = None
    _protocols: dict[Address, A2SClientProtocol]

    def __init__(self, sock: socket.socket) -> None:
        if sock.type != socket.SOCK_DGRAM:
            raise TypeError("Socket type must be SOCK_DGRAM")

        self._sock = sock
        self._remote_addr = None  # FIXME: can this be introspected from sock?
        self._protocols = {}

    # Connection methods

    def __enter__(self) -> Self:
        self._sock.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, tb) -> None:
        return self._sock.__exit__(exc_type, exc_val, tb)

    # Constructor methods

    @classmethod
    def from_addr(
        cls,
        host: str,
        port: int,
        timeout: float | None = DEFAULT_TIMEOUT,
        *,
        prefer_ipv4: bool = True,
    ) -> Self:
        """Resolve the given host and create an A2S query.

        :param host: The IPv4 address, IPv6 address, or domain name to query.
        :param port: The port to query.
        :param timeout: The timeout to set on the socket.
        :param prefer_ipv4:
            If True, prefer to resolve hostnames to IPv4 addresses.
            This may still connect the socket to an IPv6 address so if you
            need more control, consider using :func:`socket.getaddrinfo()`
            to manually create a socket and pass it to the constructor.

            .. versionadded:: 0.2.1
        :raises OSError: The address could not be resolved.

        """
        addresses = socket.getaddrinfo(
            host,
            port,
            type=socket.SOCK_DGRAM,
            proto=socket.IPPROTO_UDP,
        )
        if not addresses:
            raise OSError("Address could not be resolved")

        if prefer_ipv4:
            addr = next((a for a in addresses if a[0] == socket.AF_INET), addresses[0])
        else:
            addr = addresses[0]

        family, type, proto, _, addr = addr
        sock = socket.socket(family, type, proto)
        sock.settimeout(timeout)
        sock.connect(addr)

        obj = cls(sock)
        obj._remote_addr = addr  # Default address when user calls addr=None
        return obj

    @classmethod
    def from_ipv4(cls, timeout: float | None = DEFAULT_TIMEOUT) -> Self:
        """Create an A2S query with a UDP IPv4 socket not connected to any address.

        This allows you to use the same socket with ``addr`` arguments::

            with A2S.from_ipv4() as a2s:
                info = a2s.info(("127.0.0.1", 2303))
                info = a2s.info(("127.0.0.1", 27015))

        :param timeout: The timeout to set on the socket.

        .. versionadded:: 0.4.0

        """
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        sock.settimeout(timeout)
        return cls(sock)

    @classmethod
    def from_ipv6(cls, timeout: float | None = DEFAULT_TIMEOUT) -> Self:
        """Create an A2S query with a UDP IPv6 socket not connected to any address.

        This allows you to use the same socket with ``addr`` arguments::

            with A2S.from_ipv6() as a2s:
                info = a2s.info(("::1", 2303))
                info = a2s.info(("::1", 27015))

        :param timeout: The timeout to set on the socket.

        .. versionadded:: 0.4.0

        """
        sock = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        sock.settimeout(timeout)
        return cls(sock)

    # Request methods

    def info(self, addr: Address | None = None) -> ClientEventInfo:
        """Send an A2S_INFO request and wait for a response.

        .. versionchanged:: 0.5.0

            addr can now be a positional argument.

        :param addr:
            The address to send the request to.
            Does not apply if socket is already connected to an address,
            such as from :meth:`from_addr()`.

            .. versionadded:: 0.4.0

        :raises ChallengeError: The server sent too many challenges.
        :raises PayloadError: The server sent a malformed packet.
        :raises TimeoutError: The socket timed out.
        :raises TypeError: The addr argument was required or forbidden.

        """
        addr = self._get_addr(addr)
        proto = self._get_protocol(addr)
        return self._send(ClientEventInfo, addr, proto.info)

    def players(self, addr: Address | None = None) -> ClientEventPlayers:
        """Send an A2S_PLAYER request and wait for a response.

        .. versionchanged:: 0.5.0

            addr can now be a positional argument.

        :param addr:
            The address to send the request to.
            Does not apply if socket is already connected to an address,
            such as from :meth:`from_addr()`.

            .. versionadded:: 0.4.0

        :raises ChallengeError: The server sent too many challenges.
        :raises PayloadError: The server sent a malformed packet.
        :raises TimeoutError: The socket timed out.
        :raises TypeError: The addr argument was required or forbidden.

        """
        addr = self._get_addr(addr)
        proto = self._get_protocol(addr)
        return self._send(ClientEventPlayers, addr, proto.players)

    def rules(self, addr: Address | None = None) -> ClientEventRules:
        """Send an A2S_RULES request and wait for a response.

        .. versionchanged:: 0.5.0

            addr can now be a positional argument.

        :param addr:
            The address to send the request to.
            Does not apply if socket is already connected to an address,
            such as from :meth:`from_addr()`.

            .. versionadded:: 0.4.0

        :raises ChallengeError: The server sent too many challenges.
        :raises PayloadError: The server sent a malformed packet.
        :raises TimeoutError: The socket timed out.
        :raises TypeError: The addr argument was required or forbidden.

        """
        addr = self._get_addr(addr)
        proto = self._get_protocol(addr)
        return self._send(ClientEventRules, addr, proto.rules)

    def _get_addr(self, addr: Address | None) -> Address:
        if self._remote_addr and addr:
            raise TypeError("Socket has remote address, addr= is disallowed")
        elif not self._remote_addr and not addr:
            raise TypeError("Socket has no remote address, addr= is required")
        return addr or self._remote_addr  # type: ignore

    def _get_protocol(self, addr: Address) -> A2SClientProtocol:
        """Get the A2S protocol for the given address, creating a new one
        if it doesn't already exist.

        :param addr: The address to bind to, or None for the default protocol.

        """
        # NOTE: not thread-safe!
        proto = self._protocols.get(addr)
        if proto is not None:
            return proto

        proto = self._create_protocol()
        self._protocols[addr] = proto
        return proto

    def _create_protocol(self) -> A2SClientProtocol:
        """Create the A2S protocol to manage state.

        This method can be overridden by subclasses.

        """
        return A2SClientProtocol()

    def _send(
        self,
        t: Type[ClientEventT],
        addr: Address,
        payload: Callable[[], ClientPacket],
    ) -> ClientEventT:
        """Use the given request function to generate an outbound packet,
        and wait until the server responds with the given event type.

        This automatically handles challenge responses, re-sending the request
        and waiting again.

        :raises ChallengeError: The server sent too many challenges.
        :raises PayloadError: The server sent a malformed packet.
        :raises TimeoutError:
            The socket timed out, or the server did not respond with the event.

        """
        types = (t, ClientEventChallenge)

        for _ in range(3):
            self._sock.sendto(bytes(payload()), addr)
            events = list(filter_type(types, self._recv(addr)))
            if not events:
                # FIXME: not really a timeout, should be a custom exception
                raise TimeoutError(f"Server failed to respond with {t.__name__}")
            elif (found := first(t, events)) is not None:
                return found

        raise ChallengeError("Server responded with too many challenges")

    def _recv(self, addr: Address) -> list[ClientEvent]:
        """Read one datagram from the socket and pass it to the protocol.

        This may call :meth:`~socket.socket.recvfrom()` multiple times
        until a datagram from the given address is received.

        :param addr: The address to wait for a datagram from.
        :raises PayloadError: The server sent a malformed packet.
        :raises TimeoutError: The socket timed out.

        """
        # NOTE: not thread-safe!
        data, recv_addr = self._sock.recvfrom(self.buffer_size)
        events = self._receive_datagram(data, recv_addr)

        while not events or addr != recv_addr:
            data, recv_addr = self._sock.recvfrom(self.buffer_size)
            events = self._receive_datagram(data, recv_addr)

        return events

    def _receive_datagram(self, data: bytes, addr: Address) -> list[ClientEvent]:
        """Pass the datagram to the protocol and return any generated events.

        :raises PayloadError: The server sent a malformed packet.
        :raises TimeoutError: The socket timed out.

        """
        proto = self._protocols.get(addr)
        if proto is None:
            log.debug("Ignoring data from unexpected address %s", addr)
            return []

        proto.receive_datagram(data)

        return proto.events_received()


class A2SGoldsource(A2S):
    """A synchronous client for A2S Goldsource queries."""

    def info(self, addr: Address | None = None) -> ClientEventGoldsourceInfo:  # type: ignore
        addr = self._get_addr(addr)
        proto = self._get_protocol(addr)
        return self._send(ClientEventGoldsourceInfo, addr, proto.info)

    def _create_protocol(self) -> A2SGoldsourceClientProtocol:
        return A2SGoldsourceClientProtocol()



from .client import (
    A2S as A2S,
    A2SGoldsource as A2SGoldsource,
    AsyncA2S as AsyncA2S,
    AsyncA2SGoldsource as AsyncA2SGoldsource,
    filter_type as filter_type,
    first as first,
    last as last,
)
from .errors import (
    ChallengeError as ChallengeError,
    Error as Error,
    PayloadError as PayloadError,
)
from .events import (
    ClientEvent as ClientEvent,
    ClientEventChallenge as ClientEventChallenge,
    ClientEventGoldsourceInfo as ClientEventGoldsourceInfo,
    ClientEventInfo as ClientEventInfo,
    ClientEventPlayers as ClientEventPlayers,
    ClientEventRules as ClientEventRules,
    Environment as Environment,
    Event as Event,
    ExtraInfo as ExtraInfo,
    GoldsourceMod as GoldsourceMod,
    GoldsourceModDLL as GoldsourceModDLL,
    GoldsourceModType as GoldsourceModType,
    Player as Player,
    ServerType as ServerType,
    VAC as VAC,
    Visibility as Visibility,
)
from .headers import (
    Compression as Compression,
    Header as Header,
    HeaderType as HeaderType,
    MultiGoldsourceHeader as MultiGoldsourceHeader,
    MultiHeader as MultiHeader,
    SimpleHeader as SimpleHeader,
)
from .packets import (
    ClientPacket as ClientPacket,
    ClientPacketInfo as ClientPacketInfo,
    ClientPacketPlayers as ClientPacketPlayers,
    ClientPacketRules as ClientPacketRules,
    Packet as Packet,
)
from .protocol import (
    A2SClientProtocol as A2SClientProtocol,
    A2SGoldsourceClientProtocol as A2SGoldsourceClientProtocol,
    MultiPartResponse as MultiPartResponse,
)
from .reader import Readable as Readable, Reader as Reader
from .rules import (
    Arma3DLC as Arma3DLC,
    Arma3Difficulty as Arma3Difficulty,
    Arma3Mod as Arma3Mod,
    Arma3Rules as Arma3Rules,
)


def _get_version() -> str:
    from importlib.metadata import version

    return version(_dist_name)


_dist_name = "little-a2s"
__version__ = _get_version()
